# -*- coding: utf-8 -*-

import logging
import time

from . import atomicwrite


logger = logging.getLogger(__name__)


async def output_status_influx(config, data):
    '''Outputs the status as InfluxDB wire protocol'''
    # https://github.com/influxdata/telegraf/blob/master/docs/DATA_FORMATS_INPUT.md
    # https://docs.influxdata.com/influxdb/v1.7/write_protocols/line_protocol_tutorial/
    filename = config.get('filename', '/var/cache/wg-track_influx.out')
    with atomicwrite.open_for_atomic_write(filename, perm=0o644) as f:
      for interface, interfacedata, peer, peerdata in data.peeriterator():
          timestamp = peerdata.get('timestamp')
          if timestamp is None:
              timestamp = time.time()
          timestamp = '{:.0f}'.format(timestamp*1000000000)
          attrs = ['transfer-rx', 'transfer-tx', 'status']
          
          readings = []
          for attr in attrs:
              value = peerdata.get(attr)
              if value is None: # None values shall not be included but these attributes just be omitted
                continue
              value = str(value) # Make sure we're dealing with strings now
              value = value.replace('=','').replace(',','').replace('"','').replace(' ','') # remove any special characters that have meaning in the Influx wire protocol
              attr = attr.replace('-', '_') # Influx convention is to use underscores instead of dashes
              if attr in ['transfer-rx', 'transfer-tx']: #integer values
                if value is not None:
                  value += 'i'
              elif attr in ['status']: #string values
                value = '"' + value + '"'
              reading = '{attr}={value}'.format(attr=attr, value=value)
              readings.append(reading)
          if peerdata.get('status') == 'up:ok':
              readings.append('is_up=1i')
          else:
              readings.append('is_up=0i')
          peer = peer.replace('=', '\=') # the equal sign needs to be escaped
          readings = ','.join(readings)    
          reading = 'wgtrack,interface={interface},peer={peer} {readings} {timestamp}\n'.format(interface=interface, peer=peer, readings=readings, timestamp=timestamp)
          f.write(reading)

async def output_status(outputs, data):
    '''Outputs the status information in the requested formats'''
    for output, output_config in outputs.items():
        if output == 'influx':
            await output_status_influx(output_config, data)
        else:
            logger.error('Unknown output [[{0}] specified in config file'.format(output))



# -*- coding: utf-8 -*-

import configparser
import logging


class Config(configparser.ConfigParser):
    '''ConfigParser with application-specific helper methods/properties defining defaults'''
    
    def __init__(self, configfile=None):
        '''Constructor'''
        super().__init__()
        if configfile is not None:
            self.read(configfile)
        
    def read(self, configfile):
        '''Read config, make sure general section exists, and remember filename'''
        super().read(configfile)
        if not self.has_section('general'):
            self.add_section('general')
        self['general']['configfile'] = configfile

    @property
    def cycle_time(self):
        return float(self['general'].get('cycle_time', 30))

    @property
    def cycles_wait(self):
        return int(self['general'].get('cycles_wait', 2))
    
    @property
    def cycles_checking(self):
        return int(self['general'].get('cycles_checking', 4))

    @property
    def cycles_checkperiod(self):
        return int(self['general'].get('cycles_checkperiod', 2))

    @property
    def cycles_slowcheckingperiod(self):
        return int(self['general'].get('cycles_slowcheckingperiod', 20))

    @property
    def ping_interval(self):
        return int(self['general'].get('ping_interval', 2))

    @property
    def ping_failafternum(self):
        return int(self['general'].get('ping_failafternum', 2))

    @property
    def loglevel(self):
        return int(self['general'].get('loglevel', logging.INFO))

    @loglevel.setter
    def loglevel(self, value):
        if value is not None:
            self['general']['loglevel'] = str(value)

    @property
    def outputs(self):
        return { k[7:]: self[k] for k in self.sections() if k.startswith('output:') }



# -*- coding: utf-8 -*-

import asyncio
import concurrent
import logging
import pprint
import signal
import socket
import sys
import time

from . import logic


logger = logging.getLogger(__name__);


class EventProcessor():
    '''Class for providing event processing for the application (uses an asyncio event loop)'''

    def __init__(self, config):
        '''Constructor'''
        self.config = config
        self.loop = asyncio.get_event_loop()
        # no longer working with Python 3.10: self.queue = asyncio.Queue(loop=self.loop)
        self.queue = asyncio.Queue(**({"loop": loop} if sys.version_info[:2] < (3, 10) else {}))
        self.logic = logic.Logic(config, self.enqueue)

    def handle_hup(self, signum, frame):
        '''Handle the SIGHUP signal'''
        logger.info('Signal "SIGHUP" received; reloading config')
        self.logic.initialize_data()

    def handle_exception(self, loop, context):
        '''Handler for exceptions in coroutines'''
        if isinstance(context.get('exception'), asyncio.CancelledError):
            logger.debug('Async event loop cancelled')
        else:
            logging.error('Caught exception: [{e}] [{m}] [{f}]'.format(e=context.get('exception', ''), m=context.get('message'), f=context.get('future')))

    def eventloop(self):
        '''Run the event processing loop'''
        #self.loop.set_debug(logger.getEffectiveLevel() <= logging.DEBUG) # in case of INFO, WARN, ERROR,... do not go into debug mode
        try:        
            self.loop.set_exception_handler(self.handle_exception)
            self.loop.run_until_complete(self.run_async())            
        finally:
            for task in asyncio.Task.all_tasks():
                task.cancel()
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())
            self.loop.close()

    async def handle_message(self, reader, writer):
        '''Process incoming messages/commands'''
        data = await reader.read(100)
        message = data.decode()
        message = message.strip()
        addr = writer.get_extra_info('peername')
        logger.info(f'Received {message!r} from {addr!r}')
        if message == 'quit':
            raise KeyboardInterrupt
        #logger.debug(f'Sending: {message!r}')
        #writer.write(data)
        #await writer.drain()
        writer.close()

    async def run_periodically(self, cycle_time):
        '''Schedules tasks periodically each "cycle_time" (interval in seconds)'''
        while True:
            start = time.time()
            await self.logic.do_periodically()
            remaining_time = cycle_time - (time.time() - start)
            if remaining_time <= 0:
                logger.warning('Periodic tasks took longer than the cycle time; increase cycle time')
            await asyncio.sleep(remaining_time)

    async def enqueue(self, command, data):
        '''Enqueues an item in the event queue'''
        await self.queue.put({'command': command, 'data': data})

    async def serve_queue(self):
        '''Serve the event queue asynchronously'''
        while True:
            item = await self.queue.get()
            await self.logic.process_queue(item)

    async def run_async(self):
        '''Asynchronously executed code'''
        # Periodic tasks
        cycle_time = self.config.cycle_time
        task_periodic = asyncio.ensure_future(self.run_periodically(cycle_time))
        # Work queue
        task_queue =  asyncio.ensure_future(self.serve_queue())
        # Listener for receiving commands
        #THIS IS WORKING BUT CURRENTLY NOT NEEDED
        #server = await asyncio.start_server(self.handle_message, '127.0.0.1', 8888)    
        #addr = server.sockets[0].getsockname()
        #print(f'Serving on {addr}')
        #task_server = asyncio.ensure_future(server.serve_forever())
        task_server = asyncio.ensure_future(asyncio.sleep(1)) # TEMPORARY REPLACEMENT
        # Wait for all tasks to finish
        await asyncio.gather(task_periodic, task_queue, task_server)


def run(config):
    '''Creates an instance and runs it'''
    evt = EventProcessor(config)
    signal.signal(signal.SIGHUP, evt.handle_hup)
    evt.eventloop()


if __name__ == '__main__':
    run()



import os
import setuptools


with open('README.md', 'r') as f:
    long_description = f.read()

setup_kwargs = {
    'name': 'wgtrack',
    'version': '0.1.10',
    'author': 'The Towalink Project',
    'author_email': 'pypi.wgtrack@towalink.net',
    'description': 'wgtrack tracks WireGuard links, exports the status, and updates endpoints as needed',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'url': 'https://www.towalink.net',
    'packages': setuptools.find_packages('src'),
    'package_dir': {'': 'src'},
    'include_package_data': True,
    'install_requires': [ ],
    'entry_points': '''
        [console_scripts]
        wgtrack=wgtrack:main
    ''',
    'classifiers': [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: POSIX :: Linux',
        #'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Telecommunications Industry',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Networking'
    ],
    'python_requires': '>=3.5',
    'keywords': 'WireGuard monitoring Towalink VPN dyndns NAT-traversal',
    'project_urls': {
        'Repository': 'https://www.github.com/towalink/wgtrack',
    },
}


if __name__ == '__main__':
    setuptools.setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import textwrap

from . import exechelper


def is_root():
    """Returns whether this script is run with user id 0 (root)"""
    return os.getuid() == 0
    
def check_wg():
    """Check whether the wg tool is present"""
    return os.path.isfile('/usr/bin/wg')

def get_startupscript_openrc():
    """Get an openrc startup script for wgtrack"""
    template = r'''
        #!/sbin/openrc-run
        
        depend() {
            need net
        }
        
        name=$RC_SVCNAME
        command="wgtrack"
        command_args=""
        pidfile="/run/$RC_SVCNAME.pid"
        command_background="yes"
        stopsig="SIGTERM"
        start_stop_daemon_args="--stdout /var/log/wgtrack.log --stderr /var/log/wgtrack.err"
        '''
    template = textwrap.dedent(template).lstrip()
    return template

def get_startupscript_systemd():
    """Get a systemd service for wgtrack"""
    template = r'''
        [Unit]
        Description=wgtrack service
        After=network-online.target
        Wants=network-online.target
        
        [Service]
        Type=simple
        ExecStart=wgtrack
        StandardOutput=append:/var/log/wgtrack.log
        StandardError=inherit
        
        [Install]
        WantedBy=multi-user.target
        '''
    template = textwrap.dedent(template).lstrip()
    return template

def install_startupscript_openrc():
    """Install a startup script for wgtrack to /etc/init.d"""
    filename = '/etc/init.d/wgtrack'
    with open(filename, 'w') as f:
        f.write(get_startupscript_openrc())
    os.chmod(filename, 0o700)

def install_startupscript_systemd(eh):
    """Install a systemd service for wgtrack"""
    if os.path.isdir('/etc/systemd/system'):
        with open('/etc/systemd/system/wgtrack.service', 'w') as systemd_file:
            systemd_file.write(get_startupscript_systemd())  
        eh.execute('systemctl daemon-reload', suppressoutput=True, suppresserrors=True)
        return True
    else:
        return False

def setup_environment(install=True):
    """Environment setup assistant"""
    if not is_root():
        print('Error: this needs to be done as root user. Aborting.')
        exit(1)
    eh = exechelper.ExecHelper()
    if install:
        if not check_wg():
            print('Error: Wireguard (wg) is not available on this system. Aborting.')
            exit(1)
        if eh.os_id == 'alpine':
            # Create and install openrc startup script
            install_startupscript_openrc()
        else:
            # Create and install systemd service
            if not install_startupscript_systemd(eh):
                print('Error: this is neither running on Alpine nor has systemd been found')
                exit(1)
        eh.start_service('wgtrack')
        eh.enable_service('wgtrack')
    else:    
        eh.disable_service('wgtrack')
        eh.stop_service('wgtrack')
        try:
            if eh.os_id == 'alpine':
                os.remove('/etc/init.d/wgtrack')
            else:
                os.remove('/etc/systemd/system/wgtrack.service')
                eh.execute('systemctl daemon-reload', suppressoutput=True, suppresserrors=True)
        except FileNotFoundError:
            pass


if __name__ == '__main__':
    setup_environment()
